"use strict";
const EntityMetadata_1 = require("../metadata/EntityMetadata");
const ColumnMetadata_1 = require("../metadata/ColumnMetadata");
const ForeignKeyMetadata_1 = require("../metadata/ForeignKeyMetadata");
const TableMetadata_1 = require("../metadata/TableMetadata");
/**
 * Helps to create EntityMetadatas for junction tables.
 *
 * @internal
 */
class JunctionEntityMetadataBuilder {
    build(lazyRelationsWrapper, args) {
        const column1 = args.joinTable.referencedColumn;
        const column2 = args.joinTable.inverseReferencedColumn;
        const tableMetadata = new TableMetadata_1.TableMetadata({
            name: args.joinTable.name,
            type: "junction"
        });
        const junctionColumn1 = new ColumnMetadata_1.ColumnMetadata({
            propertyType: column1.type,
            mode: "virtual",
            options: {
                length: column1.length,
                type: column1.type,
                name: args.joinTable.joinColumnName
            }
        });
        const junctionColumn2 = new ColumnMetadata_1.ColumnMetadata({
            propertyType: column2.type,
            mode: "virtual",
            options: {
                length: column2.length,
                type: column2.type,
                name: args.joinTable.inverseJoinColumnName
            }
        });
        return new EntityMetadata_1.EntityMetadata(lazyRelationsWrapper, {
            namingStrategy: args.namingStrategy,
            tableMetadata: tableMetadata,
            columnMetadatas: [
                junctionColumn1,
                junctionColumn2
            ],
            foreignKeyMetadatas: [
                new ForeignKeyMetadata_1.ForeignKeyMetadata([junctionColumn1], args.firstTable, [column1]),
                new ForeignKeyMetadata_1.ForeignKeyMetadata([junctionColumn2], args.secondTable, [column2])
            ]
        });
    }
}
exports.JunctionEntityMetadataBuilder = JunctionEntityMetadataBuilder;
//# sourceMappingURL=JunctionEntityMetadataBuilder.js.map