"use strict";
const PropertyMetadata_1 = require("./PropertyMetadata");
/**
 * JoinColumnMetadata contains all information about relation's join column.
 */
class JoinColumnMetadata extends PropertyMetadata_1.PropertyMetadata {
    // ---------------------------------------------------------------------
    // Constructor
    // ---------------------------------------------------------------------
    constructor(args) {
        super(args.target, args.propertyName);
        this._name = args.name;
        this.referencedColumnName = args.referencedColumnName;
    }
    // ---------------------------------------------------------------------
    // Accessors
    // ---------------------------------------------------------------------
    /**
     * Join column name.
     */
    get name() {
        return this.relation.entityMetadata.namingStrategy.joinColumnInverseSideName(this._name, this.relation.propertyName);
    }
    /**
     * Referenced join column.
     */
    get referencedColumn() {
        if (this.referencedColumnName) {
            const referencedColumn = this.relation.inverseEntityMetadata.columns.find(column => column.name === this.referencedColumnName);
            if (!referencedColumn)
                throw new Error(`Referenced column ${this.referencedColumnName} was not found in entity ${this.name}`);
        }
        return this.relation.inverseEntityMetadata.primaryColumn;
    }
}
exports.JoinColumnMetadata = JoinColumnMetadata;
//# sourceMappingURL=JoinColumnMetadata.js.map