"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const Alias_1 = require("./alias/Alias");
const AliasMap_1 = require("./alias/AliasMap");
const RawSqlResultsToEntityTransformer_1 = require("./transformer/RawSqlResultsToEntityTransformer");
class QueryBuilder {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    constructor(driver, entityMetadatas, broadcaster) {
        this.driver = driver;
        this.entityMetadatas = entityMetadatas;
        this.broadcaster = broadcaster;
        this.type = "select";
        this.selects = [];
        this.joins = [];
        this.joinRelationIds = [];
        this.relationCountMetas = [];
        this.groupBys = [];
        this.wheres = [];
        this.havings = [];
        this.orderBys = [];
        this.parameters = {};
        this.aliasMap = new AliasMap_1.AliasMap(entityMetadatas);
    }
    // -------------------------------------------------------------------------
    // Accessors
    // -------------------------------------------------------------------------
    get alias() {
        return this.aliasMap.mainAlias.name;
    }
    delete(tableNameOrEntity) {
        if (tableNameOrEntity instanceof Function) {
            const aliasName = tableNameOrEntity.name;
            const aliasObj = new Alias_1.Alias(aliasName);
            aliasObj.target = tableNameOrEntity;
            this.aliasMap.addMainAlias(aliasObj);
            this.fromEntity = { alias: aliasObj };
        }
        else if (typeof tableNameOrEntity === "string") {
            this.fromTableName = tableNameOrEntity;
        }
        this.type = "delete";
        return this;
    }
    update(tableNameOrEntityOrUpdateSet, maybeUpdateSet) {
        const updateSet = maybeUpdateSet ? maybeUpdateSet : tableNameOrEntityOrUpdateSet;
        if (tableNameOrEntityOrUpdateSet instanceof Function) {
            const aliasName = tableNameOrEntityOrUpdateSet.name;
            const aliasObj = new Alias_1.Alias(aliasName);
            aliasObj.target = tableNameOrEntityOrUpdateSet;
            this.aliasMap.addMainAlias(aliasObj);
            this.fromEntity = { alias: aliasObj };
        }
        else if (typeof tableNameOrEntityOrUpdateSet === "string") {
            this.fromTableName = tableNameOrEntityOrUpdateSet;
        }
        this.type = "update";
        this.updateQuerySet = updateSet;
        return this;
    }
    select(selection) {
        this.type = "select";
        if (selection) {
            if (selection instanceof Array) {
                this.selects = selection;
            }
            else {
                this.selects = [selection];
            }
        }
        return this;
    }
    addSelect(selection) {
        if (selection instanceof Array)
            this.selects = this.selects.concat(selection);
        else
            this.selects.push(selection);
        return this;
    }
    from(entityOrTableName, alias) {
        const aliasObj = new Alias_1.Alias(alias);
        aliasObj.target = entityOrTableName;
        this.aliasMap.addMainAlias(aliasObj);
        this.fromEntity = { alias: aliasObj };
        return this;
        /* if (entityOrTableName instanceof Function) {
             const aliasObj = new Alias(alias);
             aliasObj.target = <Function> entityOrTableName;
             this.aliasMap.addMainAlias(aliasObj);
             this.fromEntity = { alias: aliasObj };
         } else {
             this.fromTableName = <string> entityOrTableName;
             this.fromTableAlias = alias;
         }
         return this;*/
    }
    countRelationAndMap(mapProperty, property, conditionType = "ON", condition = "", parameters) {
        const [parentAliasName, parentPropertyName] = property.split(".");
        const alias = parentAliasName + "_" + parentPropertyName + "_relation_count";
        const aliasObj = new Alias_1.Alias(alias);
        this.aliasMap.addAlias(aliasObj);
        aliasObj.parentAliasName = parentAliasName;
        aliasObj.parentPropertyName = parentPropertyName;
        const relationCountMeta = {
            mapToProperty: mapProperty,
            conditionType: conditionType,
            condition: condition,
            alias: aliasObj,
            entities: []
        };
        this.relationCountMetas.push(relationCountMeta);
        if (parameters)
            this.addParameters(parameters);
        return this;
    }
    countRelation(property, conditionType = "ON", condition = "", parameters) {
        const [parentAliasName, parentPropertyName] = property.split(".");
        const alias = parentAliasName + "_" + parentPropertyName + "_relation_count";
        const aliasObj = new Alias_1.Alias(alias);
        this.aliasMap.addAlias(aliasObj);
        aliasObj.parentAliasName = parentAliasName;
        aliasObj.parentPropertyName = parentPropertyName;
        const relationCountMeta = {
            conditionType: conditionType,
            condition: condition,
            alias: aliasObj,
            entities: []
        };
        this.relationCountMetas.push(relationCountMeta);
        if (parameters)
            this.addParameters(parameters);
        return this;
    }
    leftJoinRelationIdAndMap(mapToProperty, property, conditionType = "ON", condition = "", parameters) {
        return this.joinRelationId("INNER", mapToProperty, property, conditionType, condition, parameters);
    }
    innerJoinRelationIdAndMap(mapToProperty, property, conditionType = "ON", condition = "", parameters) {
        return this.joinRelationId("INNER", mapToProperty, property, conditionType, condition, parameters);
    }
    innerJoinRelationId(property, conditionType, condition, parameters) {
        return this.joinRelationId("INNER", undefined, property, conditionType, condition, parameters);
    }
    leftJoinRelationId(property, conditionType = "ON", condition, parameters) {
        return this.joinRelationId("LEFT", undefined, property, conditionType, condition, parameters);
    }
    innerJoinAndMapMany(mapToProperty, entityOrProperty, alias, conditionType = "ON", condition = "", parameters) {
        this.addSelect(alias);
        return this.join("INNER", entityOrProperty, alias, conditionType, condition, parameters, mapToProperty, true);
    }
    innerJoinAndMapOne(mapToProperty, entityOrProperty, alias, conditionType = "ON", condition = "", parameters) {
        this.addSelect(alias);
        return this.join("INNER", entityOrProperty, alias, conditionType, condition, parameters, mapToProperty, false);
    }
    innerJoinAndSelect(entityOrProperty, alias, conditionType = "ON", condition = "", parameters) {
        this.addSelect(alias);
        return this.join("INNER", entityOrProperty, alias, conditionType, condition, parameters);
    }
    innerJoin(entityOrProperty, alias, conditionType = "ON", condition = "", parameters) {
        return this.join("INNER", entityOrProperty, alias, conditionType, condition, parameters);
    }
    leftJoinAndMapMany(mapToProperty, entityOrProperty, alias, conditionType = "ON", condition = "", parameters) {
        this.addSelect(alias);
        return this.join("LEFT", entityOrProperty, alias, conditionType, condition, parameters, mapToProperty, true);
    }
    leftJoinAndMapOne(mapToProperty, entityOrProperty, alias, conditionType = "ON", condition = "", parameters) {
        this.addSelect(alias);
        return this.join("LEFT", entityOrProperty, alias, conditionType, condition, parameters, mapToProperty, false);
    }
    leftJoinAndSelect(entityOrProperty, alias, conditionType = "ON", condition = "", parameters) {
        this.addSelect(alias);
        return this.join("LEFT", entityOrProperty, alias, conditionType, condition, parameters);
    }
    leftJoin(entityOrProperty, alias, conditionType = "ON", condition = "", parameters) {
        return this.join("LEFT", entityOrProperty, alias, conditionType, condition, parameters);
    }
    where(where, parameters) {
        this.wheres.push({ type: "simple", condition: where });
        if (parameters)
            this.addParameters(parameters);
        return this;
    }
    andWhere(where, parameters) {
        this.wheres.push({ type: "and", condition: where });
        if (parameters)
            this.addParameters(parameters);
        return this;
    }
    orWhere(where, parameters) {
        this.wheres.push({ type: "or", condition: where });
        if (parameters)
            this.addParameters(parameters);
        return this;
    }
    groupBy(groupBy) {
        this.groupBys = [groupBy];
        return this;
    }
    addGroupBy(groupBy) {
        this.groupBys.push(groupBy);
        return this;
    }
    having(having, parameters) {
        this.havings.push({ type: "simple", condition: having });
        if (parameters)
            this.addParameters(parameters);
        return this;
    }
    andHaving(having, parameters) {
        this.havings.push({ type: "and", condition: having });
        if (parameters)
            this.addParameters(parameters);
        return this;
    }
    orHaving(having, parameters) {
        this.havings.push({ type: "or", condition: having });
        if (parameters)
            this.addParameters(parameters);
        return this;
    }
    orderBy(sort, order = "ASC") {
        this.orderBys = [{ sort: sort, order: order }];
        return this;
    }
    addOrderBy(sort, order = "ASC") {
        this.orderBys.push({ sort: sort, order: order });
        return this;
    }
    setLimit(limit) {
        this.limit = limit;
        return this;
    }
    setOffset(offset) {
        this.offset = offset;
        return this;
    }
    setFirstResult(firstResult) {
        this.firstResult = firstResult;
        return this;
    }
    setMaxResults(maxResults) {
        this.maxResults = maxResults;
        return this;
    }
    setParameter(key, value) {
        this.parameters[key] = value;
        return this;
    }
    setParameters(parameters) {
        this.parameters = {};
        Object.keys(parameters).forEach(key => this.parameters[key] = parameters[key]);
        return this;
    }
    addParameters(parameters) {
        Object.keys(parameters).forEach(key => this.parameters[key] = parameters[key]);
        return this;
    }
    getSql() {
        let sql = this.createSelectExpression();
        sql += this.createJoinExpression();
        sql += this.createJoinRelationIdsExpression();
        sql += this.createWhereExpression();
        sql += this.createGroupByExpression();
        sql += this.createHavingExpression();
        sql += this.createOrderByExpression();
        sql += this.createLimitExpression();
        sql += this.createOffsetExpression();
        sql = this.replaceParameters(sql);
        return sql;
    }
    execute() {
        return this.driver.query(this.getSql());
    }
    getScalarResults() {
        return this.driver.query(this.getSql());
    }
    getSingleScalarResult() {
        return this.getScalarResults().then(results => results[0]);
    }
    getResults() {
        return this.getResultsAndScalarResults().then(results => {
            return results.entities;
        });
    }
    getResultsAndScalarResults() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.aliasMap.hasMainAlias)
                throw new Error(`Alias is not set. Looks like nothing is selected. Use select*, delete, update method to set an alias.`);
            const mainAliasName = this.aliasMap.mainAlias.name;
            let scalarResults;
            if (this.firstResult || this.maxResults) {
                const metadata = this.entityMetadatas.findByTarget(this.fromEntity.alias.target);
                let idsQuery = `SELECT DISTINCT(distinctAlias.${mainAliasName}_${metadata.primaryColumn.name}) as ids`;
                if (this.orderBys && this.orderBys.length > 0)
                    idsQuery += ", " + this.orderBys.map(orderBy => orderBy.sort.replace(".", "_")).join(", ");
                idsQuery += ` FROM (${this.getSql()}) distinctAlias`;
                if (this.orderBys && this.orderBys.length > 0)
                    idsQuery += " ORDER BY " + this.orderBys.map(order => "distinctAlias." + order.sort.replace(".", "_") + " " + order.order).join(", ");
                if (this.maxResults)
                    idsQuery += " LIMIT " + this.maxResults;
                if (this.firstResult)
                    idsQuery += " OFFSET " + this.firstResult;
                return this.driver
                    .query(idsQuery)
                    .then((results) => {
                    scalarResults = results;
                    const ids = results.map(result => result["ids"]).join(", ");
                    if (ids.length === 0)
                        return Promise.resolve([]);
                    const queryWithIds = this.clone()
                        .andWhere(mainAliasName + "." + metadata.primaryColumn.name + " IN (" + ids + ")")
                        .getSql();
                    return this.driver.query(queryWithIds);
                })
                    .then(results => this.rawResultsToEntities(results))
                    .then(results => this.broadcaster.broadcastLoadEventsForAll(this.aliasMap.mainAlias.target, results).then(() => results))
                    .then(results => {
                    return {
                        entities: results,
                        scalarResults: scalarResults
                    };
                });
            }
            else {
                return this.driver
                    .query(this.getSql())
                    .then(results => {
                    scalarResults = results;
                    return this.rawResultsToEntities(results);
                })
                    .then(results => {
                    return this.loadRelationCounts(results)
                        .then(counts => {
                        // console.log("counts: ", counts);
                        return results;
                    });
                })
                    .then(results => {
                    return this.broadcaster
                        .broadcastLoadEventsForAll(this.aliasMap.mainAlias.target, results)
                        .then(() => results);
                })
                    .then(results => {
                    return {
                        entities: results,
                        scalarResults: scalarResults
                    };
                });
            }
        });
    }
    /*private extractIdMap(alias: Alias, results: Entity[]): { alias: Alias, ids: any[] }[] {

        // console.log("results ids:", results.map(result => result[]));
        const metadata = this.aliasMap.getEntityMetadataByAlias(alias);
        if (!metadata)
            throw new Error("Cannot get entity metadata for the given alias " + this.aliasMap.mainAlias.name);

        /!*
        results.forEach(result => {
        });

        metadata.relations.forEach(relation => {
            this.aliasMap.
        });
        *!/
    }*/
    loadRelationCounts(results) {
        const promises = this.relationCountMetas.map(relationCountMeta => {
            const parentAlias = relationCountMeta.alias.parentAliasName;
            const foundAlias = this.aliasMap.findAliasByName(parentAlias);
            if (!foundAlias)
                throw new Error(`Alias "${parentAlias}" was not found`);
            const parentMetadata = this.aliasMap.getEntityMetadataByAlias(foundAlias);
            if (!parentMetadata)
                throw new Error("Cannot get entity metadata for the given alias " + foundAlias.name);
            const relation = parentMetadata.findRelationWithPropertyName(relationCountMeta.alias.parentPropertyName);
            const queryBuilder = new this.constructor(this.driver, this.entityMetadatas, this.broadcaster);
            let condition = "";
            const metadata = this.aliasMap.getEntityMetadataByAlias(relationCountMeta.alias);
            if (!metadata)
                throw new Error("Cannot get entity metadata for the given alias " + relationCountMeta.alias.name);
            let joinTableName = metadata.table.name;
            const junctionMetadata = relation.junctionEntityMetadata;
            const appendedCondition = relationCountMeta.condition ? " AND " + this.replacePropertyNames(relationCountMeta.condition) : "";
            /*if (relation.isManyToMany) {
                const junctionTable = junctionMetadata.table.name;
                const junctionAlias = relationCountMeta.alias.parentAliasName + "_" + relationCountMeta.alias.name;
                const joinAlias = relationCountMeta.alias.name;
                const joinTable = relation.isOwning ? relation.joinTable : relation.inverseRelation.joinTable; // not sure if this is correct
                const joinTableColumn = joinTable.referencedColumn.name; // not sure if this is correct
                const inverseJoinColumnName = joinTable.inverseReferencedColumn.name; // not sure if this is correct

                let condition1 = "", condition2 = "";
                if (relation.isOwning) {
                    condition1 = junctionAlias + "." + junctionMetadata.columns[0].name + "=" + parentAlias + "." + joinTableColumn;
                    condition2 = joinAlias + "." + inverseJoinColumnName + "=" + junctionAlias + "." + junctionMetadata.columns[1].name;
                } else {
                    condition1 = junctionAlias + "." + junctionMetadata.columns[1].name + "=" + parentAlias + "." + joinTableColumn;
                    condition2 = joinAlias + "." + inverseJoinColumnName + "=" + junctionAlias + "." + junctionMetadata.columns[0].name;
                }

                condition = " LEFT JOIN " + junctionTable + " " + junctionAlias + " " + relationCountMeta.conditionType + " " + condition1 +
                    " LEFT JOIN " + joinTableName + " " + joinAlias + " " + relationCountMeta.conditionType + " " + condition2 + appendedCondition;

            } else if (relation.isManyToOne || (relation.isOneToOne && relation.isOwning)) {
                const joinTableColumn = relation.joinColumn.referencedColumn.name;
                const condition2 = relationCountMeta.alias.name + "." + joinTableColumn + "=" + parentAlias + "." + relation.name;
                condition = " LEFT JOIN " + joinTableName + " " + relationCountMeta.alias.name + " " + relationCountMeta.conditionType + " " + condition2 + appendedCondition;

            } else {
                throw new Error(`Relation count can be applied only `); // this should be done on entity build
            }*/
            // if (relationCountMeta.condition)
            //     condition += relationCountMeta.condition;
            // relationCountMeta.alias.target;
            const ids = relationCountMeta.entities.map(entityWithMetadata => entityWithMetadata.entity[entityWithMetadata.metadata.primaryColumn.propertyName]);
            if (!ids || !ids.length)
                throw new Error(`No ids found to load relation counters`);
            return queryBuilder
                .select(`${parentMetadata.name + "." + parentMetadata.primaryColumn.propertyName} as id`)
                .addSelect(`COUNT(${relation.propertyName + "." + relation.inverseEntityMetadata.primaryColumn.propertyName}) as cnt`)
                .from(parentMetadata.target, parentMetadata.name)
                .leftJoin(parentMetadata.name + "." + relation.propertyName, relation.propertyName, relationCountMeta.conditionType, relationCountMeta.condition)
                .setParameters(this.parameters)
                .where(`${parentMetadata.name + "." + parentMetadata.primaryColumn.propertyName} IN (:relationCountIds)`, { relationCountIds: ids })
                .groupBy(parentMetadata.name + "." + parentMetadata.primaryColumn.propertyName)
                .getScalarResults()
                .then((results) => {
                // console.log(relationCountMeta.entities);
                relationCountMeta.entities.forEach(entityWithMetadata => {
                    const entityId = entityWithMetadata.entity[entityWithMetadata.metadata.primaryColumn.propertyName];
                    const entityResult = results.find(result => result.id === entityId);
                    if (entityResult) {
                        if (relationCountMeta.mapToProperty) {
                            const [parentName, propertyName] = relationCountMeta.mapToProperty.split(".");
                            // todo: right now mapping is working only on the currently countRelation class, but 
                            // different properties are working. make different classes to work too
                            entityWithMetadata.entity[propertyName] = entityResult.cnt;
                        }
                        else if (relation.countField) {
                            entityWithMetadata.entity[relation.countField] = entityResult.cnt;
                        }
                    }
                });
            });
        });
        return Promise.all(promises);
    }
    getSingleResult() {
        return this.getResults().then(entities => entities[0]);
    }
    getCount() {
        const mainAlias = this.aliasMap.mainAlias.name;
        const metadata = this.entityMetadatas.findByTarget(this.fromEntity.alias.target);
        const countQuery = this.clone({ skipOrderBys: true })
            .select(`COUNT(DISTINCT(${mainAlias}.${metadata.primaryColumn.name})) as cnt`)
            .getSql();
        return this.driver
            .query(countQuery)
            .then(results => {
            if (!results || !results[0] || !results[0]["cnt"])
                return 0;
            return parseInt(results[0]["cnt"]);
        });
    }
    getResultsAndCount() {
        return Promise.all([
            this.getResults(),
            this.getCount()
        ]);
    }
    clone(options) {
        const qb = new QueryBuilder(this.driver, this.entityMetadatas, this.broadcaster);
        switch (this.type) {
            case "select":
                qb.select(this.selects);
                break;
            case "update":
                qb.update(this.updateQuerySet);
                break;
            case "delete":
                qb.delete();
                break;
        }
        if (this.fromEntity && this.fromEntity.alias && this.fromEntity.alias.target) {
            qb.from(this.fromEntity.alias.target, this.fromEntity.alias.name);
        }
        else if (this.fromTableName) {
            qb.from(this.fromTableName, this.fromTableAlias);
        }
        this.joins.forEach(join => {
            const property = join.tableName || join.alias.target || (join.alias.parentAliasName + "." + join.alias.parentPropertyName);
            qb.join(join.type, property, join.alias.name, join.conditionType, join.condition || "", undefined, join.mapToProperty, join.isMappingMany);
        });
        this.groupBys.forEach(groupBy => qb.addGroupBy(groupBy));
        this.wheres.forEach(where => {
            switch (where.type) {
                case "simple":
                    qb.where(where.condition);
                    break;
                case "and":
                    qb.andWhere(where.condition);
                    break;
                case "or":
                    qb.orWhere(where.condition);
                    break;
            }
        });
        this.havings.forEach(having => {
            switch (having.type) {
                case "simple":
                    qb.having(having.condition);
                    break;
                case "and":
                    qb.andHaving(having.condition);
                    break;
                case "or":
                    qb.orHaving(having.condition);
                    break;
            }
        });
        if (!options || !options.skipOrderBys)
            this.orderBys.forEach(orderBy => qb.addOrderBy(orderBy.sort, orderBy.order));
        Object.keys(this.parameters).forEach(key => qb.setParameter(key, this.parameters[key]));
        if (!options || !options.skipLimit)
            qb.setLimit(this.limit);
        if (!options || !options.skipOffset)
            qb.setOffset(this.offset);
        qb.setFirstResult(this.firstResult)
            .setMaxResults(this.maxResults);
        return qb;
    }
    // -------------------------------------------------------------------------
    // Protected Methods
    // -------------------------------------------------------------------------
    rawResultsToEntities(results) {
        const transformer = new RawSqlResultsToEntityTransformer_1.RawSqlResultsToEntityTransformer(this.driver, this.aliasMap, this.extractJoinMappings(), this.relationCountMetas);
        return transformer.transform(results);
    }
    createSelectExpression() {
        // todo throw exception if selects or from is missing
        let alias = "", tableName;
        const allSelects = [];
        if (this.fromTableName) {
            tableName = this.fromTableName;
            alias = this.fromTableAlias;
        }
        else if (this.fromEntity) {
            const metadata = this.aliasMap.getEntityMetadataByAlias(this.fromEntity.alias);
            if (!metadata)
                throw new Error("Cannot get entity metadata for the given alias " + this.fromEntity.alias.name);
            tableName = metadata.table.name;
            alias = this.fromEntity.alias.name;
            // add select from the main table
            if (this.selects.indexOf(alias) !== -1) {
                metadata.columns.forEach(column => {
                    allSelects.push(alias + "." + column.name + " AS " + alias + "_" + column.name);
                });
            }
        }
        else {
            throw new Error("No from given");
        }
        // add selects from joins
        this.joins
            .filter(join => this.selects.indexOf(join.alias.name) !== -1)
            .forEach(join => {
            const joinMetadata = this.aliasMap.getEntityMetadataByAlias(join.alias);
            if (joinMetadata) {
                joinMetadata.columns.forEach(column => {
                    allSelects.push(join.alias.name + "." + column.name + " AS " + join.alias.name + "_" + column.propertyName);
                });
            }
            else {
                allSelects.push(join.alias.name);
            }
        });
        // add selects from relation id joins
        this.joinRelationIds.forEach(join => {
            // const joinMetadata = this.aliasMap.getEntityMetadataByAlias(join.alias);
            const parentAlias = join.alias.parentAliasName;
            const foundAlias = this.aliasMap.findAliasByName(parentAlias);
            if (!foundAlias)
                throw new Error(`Alias "${parentAlias}" was not found`);
            const parentMetadata = this.aliasMap.getEntityMetadataByAlias(foundAlias);
            if (!parentMetadata)
                throw new Error("Cannot get entity metadata for the given alias " + foundAlias.name);
            const relation = parentMetadata.findRelationWithPropertyName(join.alias.parentPropertyName);
            const junctionMetadata = relation.junctionEntityMetadata;
            // const junctionTable = junctionMetadata.table.name;
            junctionMetadata.columns.forEach(column => {
                allSelects.push(join.alias.name + "." + column.name + " AS " + join.alias.name + "_" + column.name);
            });
        });
        // add all other selects
        this.selects.filter(select => {
            return select !== alias && !this.joins.find(join => join.alias.name === select);
        }).forEach(select => allSelects.push(select));
        // if still selection is empty, then simply set it to all (*)
        if (allSelects.length === 0)
            allSelects.push("*");
        // create a selection query
        switch (this.type) {
            case "select":
                return "SELECT " + allSelects.join(", ") + " FROM " + tableName + " " + alias;
            case "delete":
                return "DELETE " + (alias ? alias : "") + " FROM " + tableName + " " + (alias ? alias : "");
            case "update":
                const updateSet = Object.keys(this.updateQuerySet).map(key => key + "=:updateQuerySet_" + key);
                const params = Object.keys(this.updateQuerySet).reduce((object, key) => {
                    // todo: map propertyNames to names ?
                    object["updateQuerySet_" + key] = this.updateQuerySet[key];
                    return object;
                }, {});
                this.addParameters(params);
                return "UPDATE " + tableName + " " + (alias ? alias : "") + " SET " + updateSet;
        }
        throw new Error("No query builder type is specified.");
    }
    createWhereExpression() {
        if (!this.wheres || !this.wheres.length)
            return "";
        return " WHERE " + this.wheres.map((where, index) => {
            switch (where.type) {
                case "and":
                    return (index > 0 ? "AND " : "") + this.replacePropertyNames(where.condition);
                case "or":
                    return (index > 0 ? "OR " : "") + this.replacePropertyNames(where.condition);
                default:
                    return this.replacePropertyNames(where.condition);
            }
        }).join(" ");
    }
    /**
     * Replaces all entity's propertyName to name in the given statement.
     */
    replacePropertyNames(statement) {
        this.aliasMap.aliases.forEach(alias => {
            const metadata = this.aliasMap.getEntityMetadataByAlias(alias);
            if (!metadata)
                return;
            metadata.columns.forEach(column => {
                statement = statement.replace(new RegExp(alias.name + "." + column.propertyName, "g"), alias.name + "." + column.name);
            });
            metadata.relationsWithJoinColumns.forEach(relation => {
                statement = statement.replace(new RegExp(alias.name + "." + relation.propertyName, "g"), alias.name + "." + relation.name);
            });
        });
        return statement;
    }
    createJoinRelationIdsExpression() {
        return this.joinRelationIds.map(join => {
            const parentAlias = join.alias.parentAliasName;
            const foundAlias = this.aliasMap.findAliasByName(parentAlias);
            if (!foundAlias)
                throw new Error(`Alias "${parentAlias}" was not found`);
            const parentMetadata = this.aliasMap.getEntityMetadataByAlias(foundAlias);
            if (!parentMetadata)
                throw new Error("Cannot get entity metadata for the given alias " + foundAlias.name);
            const relation = parentMetadata.findRelationWithPropertyName(join.alias.parentPropertyName);
            const junctionMetadata = relation.junctionEntityMetadata;
            const junctionTable = junctionMetadata.table.name;
            const junctionAlias = join.alias.name;
            const joinTable = relation.isOwning ? relation.joinTable : relation.inverseRelation.joinTable; // not sure if this is correct
            const joinTableColumn = joinTable.referencedColumn.name; // not sure if this is correct
            let condition1 = "";
            if (relation.isOwning) {
                condition1 = junctionAlias + "." + junctionMetadata.columns[0].name + "=" + parentAlias + "." + joinTableColumn;
            }
            else {
                condition1 = junctionAlias + "." + junctionMetadata.columns[1].name + "=" + parentAlias + "." + joinTableColumn;
            }
            return " " + join.type + " JOIN " + junctionTable + " " + junctionAlias + " " + join.conditionType + " " + condition1;
            // " " + joinType + " JOIN " + joinTableName + " " + joinAlias + " " + join.conditionType + " " + condition2 + appendedCondition;
            // console.log(join);
            // return " " + join.type + " JOIN " + joinTableName + " " + join.alias.name + " " + (join.condition ? (join.conditionType + " " + join.condition) : "");
        });
    }
    createJoinExpression() {
        return this.joins.map(join => {
            const joinType = join.type; // === "INNER" ? "INNER" : "LEFT";
            let joinTableName = join.tableName;
            if (!joinTableName) {
                const metadata = this.aliasMap.getEntityMetadataByAlias(join.alias);
                if (!metadata)
                    throw new Error("Cannot get entity metadata for the given alias " + join.alias.name);
                joinTableName = metadata.table.name;
            }
            const parentAlias = join.alias.parentAliasName;
            if (!parentAlias) {
                return " " + joinType + " JOIN " + joinTableName + " " + join.alias.name + " " + join.conditionType + " " + join.condition;
            }
            const foundAlias = this.aliasMap.findAliasByName(parentAlias);
            if (!foundAlias)
                throw new Error(`Alias "${parentAlias}" was not found`);
            const parentMetadata = this.aliasMap.getEntityMetadataByAlias(foundAlias);
            if (!parentMetadata)
                throw new Error("Cannot get entity metadata for the given alias " + foundAlias.name);
            const relation = parentMetadata.findRelationWithPropertyName(join.alias.parentPropertyName);
            const junctionMetadata = relation.junctionEntityMetadata;
            const appendedCondition = join.condition ? " AND " + this.replacePropertyNames(join.condition) : "";
            if (relation.isManyToMany) {
                const junctionTable = junctionMetadata.table.name;
                const junctionAlias = join.alias.parentAliasName + "_" + join.alias.name;
                const joinAlias = join.alias.name;
                const joinTable = relation.isOwning ? relation.joinTable : relation.inverseRelation.joinTable; // not sure if this is correct
                const joinTableColumn = joinTable.referencedColumn.name; // not sure if this is correct
                const inverseJoinColumnName = joinTable.inverseReferencedColumn.name; // not sure if this is correct
                let condition1 = "", condition2 = "";
                if (relation.isOwning) {
                    condition1 = junctionAlias + "." + junctionMetadata.columns[0].name + "=" + parentAlias + "." + joinTableColumn;
                    condition2 = joinAlias + "." + inverseJoinColumnName + "=" + junctionAlias + "." + junctionMetadata.columns[1].name;
                }
                else {
                    condition1 = junctionAlias + "." + junctionMetadata.columns[1].name + "=" + parentAlias + "." + joinTableColumn;
                    condition2 = joinAlias + "." + inverseJoinColumnName + "=" + junctionAlias + "." + junctionMetadata.columns[0].name;
                }
                return " " + joinType + " JOIN " + junctionTable + " " + junctionAlias + " " + join.conditionType + " " + condition1 +
                    " " + joinType + " JOIN " + joinTableName + " " + joinAlias + " " + join.conditionType + " " + condition2 + appendedCondition;
            }
            else if (relation.isManyToOne || (relation.isOneToOne && relation.isOwning)) {
                const joinTableColumn = relation.joinColumn.referencedColumn.name;
                const condition = join.alias.name + "." + joinTableColumn + "=" + parentAlias + "." + relation.name;
                return " " + joinType + " JOIN " + joinTableName + " " + join.alias.name + " " + join.conditionType + " " + condition + appendedCondition;
            }
            else if (relation.isOneToMany || (relation.isOneToOne && !relation.isOwning)) {
                const joinTableColumn = relation.inverseRelation.joinColumn.referencedColumn.name;
                const condition = join.alias.name + "." + relation.inverseRelation.name + "=" + parentAlias + "." + joinTableColumn;
                return " " + joinType + " JOIN " + joinTableName + " " + join.alias.name + " " + join.conditionType + " " + condition + appendedCondition;
            }
            else {
                throw new Error("Unexpected relation type"); // this should not be possible
            }
        }).join(" ");
    }
    createGroupByExpression() {
        if (!this.groupBys || !this.groupBys.length)
            return "";
        return " GROUP BY " + this.replacePropertyNames(this.groupBys.join(", "));
    }
    createHavingExpression() {
        if (!this.havings || !this.havings.length)
            return "";
        return " HAVING " + this.havings.map(having => {
            switch (having.type) {
                case "and":
                    return " AND " + this.replacePropertyNames(having.condition);
                case "or":
                    return " OR " + this.replacePropertyNames(having.condition);
                default:
                    return " " + this.replacePropertyNames(having.condition);
            }
        }).join(" ");
    }
    createOrderByExpression() {
        if (!this.orderBys || !this.orderBys.length)
            return "";
        return " ORDER BY " + this.orderBys.map(order => this.replacePropertyNames(order.sort) + " " + order.order).join(", ");
    }
    createLimitExpression() {
        if (!this.limit)
            return "";
        return " LIMIT " + this.limit;
    }
    createOffsetExpression() {
        if (!this.offset)
            return "";
        return " OFFSET " + this.offset;
    }
    replaceParameters(sql) {
        Object.keys(this.parameters).forEach(key => {
            const value = this.parameters[key] !== null && this.parameters[key] !== undefined ? this.driver.escape(this.parameters[key]) : "NULL";
            sql = sql.replace(new RegExp(":" + key, "g"), value); // todo: make replace only in value statements, otherwise problems
        });
        return sql;
    }
    extractJoinMappings() {
        const mappings = [];
        this.joins
            .filter(join => !!join.mapToProperty)
            .forEach(join => {
            const [parentName, propertyName] = join.mapToProperty.split(".");
            mappings.push({
                type: "join",
                alias: join.alias,
                parentName: parentName,
                propertyName: propertyName,
                isMany: join.isMappingMany
            });
        });
        this.joinRelationIds
            .filter(join => !!join.mapToProperty)
            .forEach(join => {
            const [parentName, propertyName] = join.mapToProperty.split(".");
            mappings.push({
                type: "relationId",
                alias: join.alias,
                parentName: parentName,
                propertyName: propertyName,
                isMany: false
            });
        });
        return mappings;
    }
    join(joinType, entityOrProperty, alias, conditionType = "ON", condition = "", parameters, mapToProperty, isMappingMany = false) {
        let tableName = "";
        const aliasObj = new Alias_1.Alias(alias);
        this.aliasMap.addAlias(aliasObj);
        if (entityOrProperty instanceof Function) {
            aliasObj.target = entityOrProperty;
        }
        else if (this.isPropertyAlias(entityOrProperty)) {
            [aliasObj.parentAliasName, aliasObj.parentPropertyName] = entityOrProperty.split(".");
        }
        else if (typeof entityOrProperty === "string") {
            tableName = entityOrProperty;
            if (!mapToProperty)
                mapToProperty = entityOrProperty;
        }
        const join = { type: joinType, alias: aliasObj, tableName: tableName, conditionType: conditionType, condition: condition, mapToProperty: mapToProperty, isMappingMany: isMappingMany };
        this.joins.push(join);
        if (parameters)
            this.addParameters(parameters);
        return this;
    }
    joinRelationId(joinType, mapToProperty, property, conditionType = "ON", condition, parameters) {
        if (!this.isPropertyAlias(property))
            throw new Error("Only entity relations are allowed in the leftJoinRelationId operation"); // todo: also check if that relation really has entityId
        const [parentAliasName, parentPropertyName] = property.split(".");
        const alias = parentAliasName + "_" + parentPropertyName + "_relation_id";
        const aliasObj = new Alias_1.Alias(alias);
        this.aliasMap.addAlias(aliasObj);
        aliasObj.parentAliasName = parentAliasName;
        aliasObj.parentPropertyName = parentPropertyName;
        this.joinRelationIds.push({
            type: joinType,
            mapToProperty: mapToProperty,
            alias: aliasObj,
            conditionType: conditionType,
            condition: condition
        });
        if (parameters)
            this.addParameters(parameters);
        return this;
    }
    isPropertyAlias(str) {
        if (!(typeof str === "string"))
            return false;
        if (str.indexOf(".") === -1)
            return false;
        const aliasName = str.split(".")[0];
        const propertyName = str.split(".")[1];
        if (!aliasName || !propertyName)
            return false;
        const aliasNameRegexp = /^[a-zA-Z0-9_-]+$/;
        const propertyNameRegexp = aliasNameRegexp;
        if (!aliasNameRegexp.test(aliasName) || !propertyNameRegexp.test(propertyName))
            return false;
        return true;
    }
}
exports.QueryBuilder = QueryBuilder;
//# sourceMappingURL=QueryBuilder.js.map