import { ReactiveRepository } from "./ReactiveRepository";
import { TreeRepository } from "./TreeRepository";
import { QueryBuilder } from "../query-builder/QueryBuilder";
import * as Rx from "rxjs/Rx";
/**
 * Tree repository is supposed to work with your entity objects. Find entities, insert, update, delete, etc.
 * This version of TreeRepository is using rxjs library and Observables instead of promises.
 */
export declare class ReactiveTreeRepository<Entity> extends ReactiveRepository<Entity> {
    protected repository: TreeRepository<Entity>;
    constructor(repository: TreeRepository<Entity>);
    /**
     * Roots are entities that have no ancestors. Finds them all.
     */
    findRoots(): Rx.Observable<Entity[]>;
    /**
     * Creates a query builder used to get descendants of the entities in a tree.
     */
    createDescendantsQueryBuilder(alias: string, closureTableAlias: string, entity: Entity): QueryBuilder<Entity>;
    /**
     * Gets all children (descendants) of the given entity. Returns them all in a flat array.
     */
    findDescendants(entity: Entity): Rx.Observable<Entity[]>;
    /**
     * Gets all children (descendants) of the given entity. Returns them in a tree - nested into each other.
     */
    findDescendantsTree(entity: Entity): Rx.Observable<Entity>;
    /**
     * Gets number of descendants of the entity.
     */
    countDescendants(entity: Entity): Rx.Observable<number>;
    /**
     * Creates a query builder used to get ancestors of the entities in the tree.
     */
    createAncestorsQueryBuilder(alias: string, closureTableAlias: string, entity: Entity): QueryBuilder<Entity>;
    /**
     * Gets all parents (ancestors) of the given entity. Returns them all in a flat array.
     */
    findAncestors(entity: Entity): Rx.Observable<Entity[]>;
    /**
     * Gets all parents (ancestors) of the given entity. Returns them in a tree - nested into each other.
     */
    findAncestorsTree(entity: Entity): Rx.Observable<Entity>;
    /**
     * Gets number of ancestors of the entity.
     */
    countAncestors(entity: Entity): Rx.Observable<number>;
}
